@@ -1,3 +1,8 @@
+try:
+    import modin.pandas as pd
+except ImportError:
+    raise ImportError("Please run `pip install modin`")
+
 from .grid import (
     enable,
     disable,
@@ -35,8 +40,3 @@ __all__ = [
 
 __version__ = get_versions()["version"]
 del get_versions
-
-from ._version import get_versions
-
-__version__ = get_versions()["version"]
-del get_versions
