@@ -3,6 +3,7 @@ import modin.pandas as pd
 import pandas
 import numpy as np
 import json
+import warnings
 
 from types import FunctionType
 from IPython.display import display
@@ -23,6 +24,7 @@ from traitlets import (
 from itertools import chain
 from uuid import uuid4
 from six import string_types
+from . import constants
 
 # versions of pandas prior to version 0.20.0 don't support the orient='table'
 # when calling the 'to_json' function on DataFrames.  to get around this we
@@ -63,16 +65,24 @@ class _DefaultSettings(object):
             "width": None,
         }
         self._show_toolbar = True
+        self._show_history = True
         self._precision = None  # Defer to pandas.get_option
 
     def set_grid_option(self, optname, optvalue):
         self._grid_options[optname] = optvalue
 
     def set_defaults(
-        self, show_toolbar=None, precision=None, grid_options=None, column_options=None
+        self,
+        show_toolbar=None,
+        show_history=None,
+        precision=None,
+        grid_options=None,
+        column_options=None,
     ):
         if show_toolbar is not None:
             self._show_toolbar = show_toolbar
+        if show_history is not None:
+            self._show_history = show_history
         if precision is not None:
             self._precision = precision
         if grid_options is not None:
@@ -84,6 +94,10 @@ class _DefaultSettings(object):
     def show_toolbar(self):
         return self._show_toolbar
 
+    @property
+    def show_history(self):
+        return self._show_history
+
     @property
     def grid_options(self):
         return self._grid_options
@@ -132,11 +146,14 @@ class _EventHandlers(object):
 
 defaults = _DefaultSettings()
 handlers = _EventHandlers()
-HISTORY_PREFIX = "# ---- spreadsheet transformation history ----\n"
 
 
 def set_defaults(
-    show_toolbar=None, precision=None, grid_options=None, column_options=None
+    show_toolbar=None,
+    show_history=None,
+    precision=None,
+    grid_options=None,
+    column_options=None,
 ):
     """
     Set the default modin-spreadsheet options.  The options that you can set here are the
@@ -161,6 +178,7 @@ def set_defaults(
     """
     defaults.set_defaults(
         show_toolbar=show_toolbar,
+        show_history=show_history,
         precision=precision,
         grid_options=grid_options,
         column_options=column_options,
@@ -326,6 +344,7 @@ def disable():
 def show_grid(
     data_frame,
     show_toolbar=None,
+    show_history=None,
     precision=None,
     grid_options=None,
     column_options=None,
@@ -364,6 +383,9 @@ def show_grid(
         Whether to show a toolbar with options for adding/removing rows.
         Adding/removing rows is an experimental feature which only works
         with DataFrames that have an integer index.
+    show_history : bool
+        Whether to show the cell containing the spreadsheet transformation
+        history.
     column_options : dict
         Column options that are to be applied to every column. See the
         Notes section below for more information on the available options,
@@ -477,6 +499,8 @@ def show_grid(
 
     if show_toolbar is None:
         show_toolbar = defaults.show_toolbar
+    if show_history is None:
+        show_history = defaults.show_history
     if precision is None:
         precision = defaults.precision
     if not isinstance(precision, Integral):
@@ -511,12 +535,13 @@ def show_grid(
     # create a visualization for the dataframe
     return SpreadsheetWidget(
         df=data_frame,
+        show_toolbar=show_toolbar,
+        show_history=show_history,
         precision=precision,
         grid_options=grid_options,
         column_options=column_options,
         column_definitions=column_definitions,
         row_edit_callback=row_edit_callback,
-        show_toolbar=show_toolbar,
     )
 
 
@@ -563,6 +588,8 @@ class SpreadsheetWidget(widgets.DOMWidget):
         Get/set the precision options being used by the current instance.
     show_toolbar : bool
         Get/set the show_toolbar option being used by the current instance.
+    show_history : bool
+        Get/set the show_history option being used by the current instance.
     column_options : bool
         Get/set the column options being used by the current instance.
     column_definitions : bool
@@ -614,6 +641,7 @@ class SpreadsheetWidget(widgets.DOMWidget):
     column_definitions = Dict({})
     row_edit_callback = Instance(FunctionType, sync=False, allow_none=True)
     show_toolbar = Bool(True, sync=True)
+    show_history = Bool(True, sync=True)
     id = Unicode(sync=True)
 
     def __init__(self, *args, **kwargs):
@@ -630,7 +658,7 @@ class SpreadsheetWidget(widgets.DOMWidget):
         if self.df is not None:
             self._update_df()
 
-        self._history = []
+        self._history = ["unfiltered_df = df.copy()"]
         self._view_msgs = []
         self._history_metadata_tag = "modin_spreadsheet" + str(uuid4())
         self._resetting_filters = False
@@ -644,6 +672,9 @@ class SpreadsheetWidget(widgets.DOMWidget):
     def _show_toolbar_default(self):
         return defaults.show_toolbar
 
+    def _show_history_default(self):
+        return defaults.show_history
+
     def on(self, names, handler):
         """
         Setup a handler to be called when a user interacts with the current
@@ -969,9 +1000,11 @@ class SpreadsheetWidget(widgets.DOMWidget):
         else:
             self._row_styles = {}
 
-        df_json = df.to_json(
-            None, orient="table", date_format="iso", double_precision=self.precision
-        )
+        with warnings.catch_warnings():
+            warnings.simplefilter("ignore", category=UserWarning)
+            df_json = df.to_json(
+                None, orient="table", date_format="iso", double_precision=self.precision
+            )
 
         if update_columns:
             self._interval_columns = []
@@ -1046,9 +1079,14 @@ class SpreadsheetWidget(widgets.DOMWidget):
         # and then call 'to_json' again to get a new version of the table
         # json that has interval columns replaced with text columns
         if len(self._interval_columns) > 0 or len(self._period_columns) > 0:
-            df_json = df.to_json(
-                None, orient="table", date_format="iso", double_precision=self.precision
-            )
+            with warnings.catch_warnings():
+                warnings.simplefilter("ignore", category=UserWarning)
+                df_json = df.to_json(
+                    None,
+                    orient="table",
+                    date_format="iso",
+                    double_precision=self.precision,
+                )
 
         self._df_json = df_json
 
@@ -1087,7 +1125,7 @@ class SpreadsheetWidget(widgets.DOMWidget):
                     self._df.sort_index(ascending=self._sort_ascending, inplace=True)
                     # Record sort index
                     self._record_transformation(
-                        f"# Sort index\n"
+                        f"{constants.SORT_INDEX}\n"
                         f"df.sort_index(ascending={self._sort_ascending}, inplace=True)"
                     )
                 else:
@@ -1098,7 +1136,7 @@ class SpreadsheetWidget(widgets.DOMWidget):
                     # Record sort index
                     # TODO: Fix level=level_index
                     self._record_transformation(
-                        f"# Sort index\n"
+                        f"{constants.SORT_INDEX}\n"
                         f"df.sort_index(level=level_index, ascending={self._sort_ascending}, "
                         f"inplace=True)"
                     )
@@ -1110,7 +1148,7 @@ class SpreadsheetWidget(widgets.DOMWidget):
                 )
                 # Record sort column
                 self._record_transformation(
-                    f"# Sort column\n"
+                    f"{constants.SORT_COLUMN}\n"
                     f"df.sort_values('{self._sort_field}', ascending={self._sort_ascending}, "
                     f"inplace=True)"
                 )
@@ -1130,7 +1168,7 @@ class SpreadsheetWidget(widgets.DOMWidget):
             helper_col = self._sort_field + self._sort_col_suffix
             self._record_transformation(
                 (
-                    f"#Sort mixed type column\n"
+                    f"{constants.SORT_MIXED_TYPE_COLUMN}\n"
                     f"df['{helper_col}'] = df['{self._sort_field}'].map(str)\n"
                     f"df.sort_values('{helper_col}', ascending={self._sort_ascending}, inplace=True)\n"
                     f"df.drop(columns='{helper_col}', inplace=True)"
@@ -1394,14 +1432,14 @@ class SpreadsheetWidget(widgets.DOMWidget):
                     col_series >= pd.to_datetime(filter_info["min"], unit="ms")
                 )
                 self._filter_conditions.append(
-                    f"unfiltered_df['{col_name}'] >= pd.to_datetime({filter_info['min']}, unit='ms'"
+                    f"unfiltered_df['{col_name}'] >= pd.to_datetime({filter_info['min']}, unit='ms')"
                 )
             if filter_info["max"] is not None:
                 conditions.append(
                     col_series <= pd.to_datetime(filter_info["max"], unit="ms")
                 )
                 self._filter_conditions.append(
-                    f"unfiltered_df['{col_name}'] <= pd.to_datetime({filter_info['max']}, unit='ms'"
+                    f"unfiltered_df['{col_name}'] <= pd.to_datetime({filter_info['max']}, unit='ms')"
                 )
         elif filter_info["type"] == "boolean":
             if filter_info["selected"] is not None:
@@ -1458,7 +1496,7 @@ class SpreadsheetWidget(widgets.DOMWidget):
             # Record reset filter
             # Other filters and sorts are reapplied after
             self._record_transformation(
-                f"# Reset filter\n" f"df = unfiltered_df.copy()"
+                f"{constants.RESET_FILTER}\n" f"df = unfiltered_df.copy()"
             )
         else:
             combined_condition = conditions[0]
@@ -1471,7 +1509,7 @@ class SpreadsheetWidget(widgets.DOMWidget):
                 ["(" + c + ")" for c in self._filter_conditions]
             )
             self._record_transformation(
-                f"# Filter columns\n"
+                f"{constants.FILTER_COLUMNS}\n"
                 f"df = unfiltered_df[{record_combined_condition}].copy()"
             )
             # Reset filter conditions
@@ -1490,8 +1528,10 @@ class SpreadsheetWidget(widgets.DOMWidget):
 
     def _handle_view_msg(self, widget, content, buffers=None):
         try:
-            self._view_msgs.append(content)
-            self._handle_view_msg_helper(content)
+            # Ignore UserWarnings in spreadsheet widget
+            with warnings.catch_warnings():
+                warnings.simplefilter("ignore", category=UserWarning)
+                self._handle_view_msg_helper(content)
         except Exception as e:
             self.log.error(e)
             self.log.exception("Unhandled exception while handling msg")
@@ -1500,7 +1540,7 @@ class SpreadsheetWidget(widgets.DOMWidget):
         """Handle incoming messages from the ModinSpreadsheetView"""
         if "type" not in content:
             return
-
+        self._view_msgs.append(content)
         if content["type"] == "edit_cell":
             col_info = self._columns[content["column"]]
             try:
@@ -1517,7 +1557,9 @@ class SpreadsheetWidget(widgets.DOMWidget):
                 self._df.loc[location] = val_to_set
                 # Record cell edit
                 self._record_transformation(
-                    f"# Edit cell\n" f"df.loc[{location}]={val_to_set}"
+                    f"{constants.EDIT_CELL}\n"
+                    f"df.loc[{location}]={repr(val_to_set)}\n"
+                    f"unfiltered_df.loc[{location}]={repr(val_to_set)}"
                 )
 
                 query = (
@@ -1525,6 +1567,9 @@ class SpreadsheetWidget(widgets.DOMWidget):
                     == content["unfiltered_index"]
                 )
                 self._unfiltered_df.loc[query, content["column"]] = val_to_set
+                self._update_table(
+                    triggered_by="edit_cell", fire_data_change_event=True
+                )
                 self._notify_listeners(
                     {
                         "name": "cell_edited",
@@ -1641,22 +1686,26 @@ class SpreadsheetWidget(widgets.DOMWidget):
             self._resetting_filters = False
             # Record reset filters before updating sort
             self._record_transformation(
-                ("# Reset all filters\n" "df = unfiltered_df.copy()")
+                (f"{constants.RESET_ALL_FILTERS}\n" "df = unfiltered_df.copy()")
             )
             self._update_sort()
             self._update_table(triggered_by="reset_filters")
             self._notify_listeners({"name": "filters_reset"})
         elif content["type"] == "initialize_history":
-            # Send metadata tag to frontend
-            self.send(
-                {
-                    "type": "initialize_history",
-                    "metadata_tag": self._history_metadata_tag,
-                }
-            )
-            self._update_history_cell()
+            if self.show_history:
+                # Send metadata tag to frontend
+                self.send(
+                    {
+                        "type": "initialize_history",
+                        "metadata_tag": self._history_metadata_tag,
+                    }
+                )
+                self._update_history_cell()
         elif content["type"] == "clear_history":
             self.clear_history(from_api=False)
+        elif content["type"] == "filter_history":
+            self._filter_relevant_history(persist=True)
+            self._notify_listeners({"name": "history_filtered", "source": "gui"})
 
     def _notify_listeners(self, event):
         # notify listeners at the module level
@@ -1754,9 +1803,10 @@ class SpreadsheetWidget(widgets.DOMWidget):
         )
         # Record row add
         self._record_transformation(
-            f"# Add row\n"
+            f"{constants.ADD_ROW}\n"
             f"last = df.loc[max(df.index)].copy()\n"
-            f"df.loc[last.name+1] = last.values"
+            f"df.loc[last.name+1] = last.values\n"
+            f"unfiltered_df.loc[last.name+1] = last.values"
         )
         return last.name
 
@@ -1883,7 +1933,9 @@ class SpreadsheetWidget(widgets.DOMWidget):
         self._update_table(triggered_by="remove_row")
         # Record remove rows
         self._record_transformation(
-            f"# Remove rows\n" f"df.drop({selected_names}, inplace=True)"
+            f"{constants.REMOVE_ROWS}\n"
+            f"df.drop({selected_names}, inplace=True)\n"
+            f"unfiltered_df.drop({selected_names}, inplace=True)"
         )
         return selected_names
 
@@ -1963,8 +2015,10 @@ class SpreadsheetWidget(widgets.DOMWidget):
         self._update_history_cell()
 
     def _update_history_cell(self):
+        if not self.show_history:
+            return
         cleaned_history = "\n".join(self._history)
-        cell_text = HISTORY_PREFIX + cleaned_history
+        cell_text = constants.HISTORY_PREFIX + cleaned_history
         self.send(
             {
                 "type": "update_history",
@@ -1977,7 +2031,7 @@ class SpreadsheetWidget(widgets.DOMWidget):
         return self._history
 
     def clear_history(self, from_api=True):
-        self._history = []
+        self._history = ["unfiltered_df = df.copy()"]
         source = "api" if from_api else "gui"
         self._notify_listeners({"name": "history_cleared", "source": source})
         self._update_history_cell()
@@ -2007,10 +2061,56 @@ class SpreadsheetWidget(widgets.DOMWidget):
         # Record sort index
         # After update_table to prevent resetting in progress button prematurely
         self._record_transformation(
-            f"# Reset sort\n" f"df.sort_index(ascending=True, inplace=True)"
+            f"{constants.RESET_SORT}\n" f"df.sort_index(ascending=True, inplace=True)"
         )
         source = "api" if from_api else "gui"
         self._notify_listeners({"name": "sort_reset", "source": source})
 
     def reset_filters(self):
         self.send({"type": "reset_filters"})
+
+    def filter_relevant_history(self, persist=True):
+        relevant_history = self._filter_relevant_history(persist)
+        self._notify_listeners({"name": "history_filtered", "source": "api"})
+        return relevant_history
+
+    def _filter_relevant_history(self, persist):
+        history = self.get_history()
+        relevant_history = []
+        # Whether a filter or sort can still be added
+        add_filter = True
+        add_sort = True
+        # Check history in reverse
+        for cmd in history[::-1]:
+            # Add the latest filter after a reset filter
+            if cmd.startswith(constants.FILTER_COLUMNS):
+                if not add_filter:
+                    continue
+                relevant_history.insert(0, cmd)
+                add_filter = False
+            # Add the latest sort after a reset sort
+            elif (
+                cmd.startswith(constants.SORT_COLUMN)
+                or cmd.startswith(constants.SORT_INDEX)
+                or cmd.startswith(constants.SORT_MIXED_TYPE_COLUMN)
+            ):
+                if not add_sort:
+                    continue
+                relevant_history.insert(0, cmd)
+                add_sort = False
+            # Prevent adding filter
+            elif cmd.startswith(constants.RESET_FILTER) or cmd.startswith(
+                constants.RESET_ALL_FILTERS
+            ):
+                add_filter = False
+            # Prevent adding sort
+            elif cmd.startswith(constants.RESET_SORT):
+                add_sort = False
+            # Include edit cell, remove/add row, initialization, etc.
+            else:
+                relevant_history.insert(0, cmd)
+        # Change internal state if persisting
+        if persist:
+            self._history = relevant_history
+            self._update_history_cell()
+        return relevant_history
